from typing import Any, Tuple


class Framebuffer:
    __slots__ = ['mglo', 'size', 'extra', 'old']

    def __init__(self, mglo, size):
        self.mglo = mglo  # type: Any
        self.size = size  # type: Tuple[int, int]
        self.extra = None  # type: Any

    @property
    def viewport(self) -> Tuple[int, int, int, int]:
        return self.mglo.viewport

    @viewport.setter
    def viewport(self, value):
        self.mglo.viewport = value

    def clear(self, attachment, color, viewport=None, color_mask=0xF):
        self.mglo.clear(attachment, color, viewport, color_mask)

    # def read(self, viewport=None, components=3, attachment=0, alignment=1, dtype='f1', np=False):
    #     return self.mglo.read(viewport, components, attachment, alignment, dtype, np)

    # def read_into(self, buffer, viewport=None, components=3, attachment=0, alignment=1):
    #     return self.mglo.read_into(buffer, viewport, components, attachment, alignment)

    def use(self):
        self.mglo.use()



from typing import Any

from .uniform import Uniform


class ComputeShader:
    __slots__ = ['mglo', 'uniforms', 'extra']

    def __init__(self, mglo, uniforms):
        self.mglo = mglo  # type: Any
        self.uniforms = uniforms  # type: Dict[str, Uniform]
        self.extra = None  # type: Any

    def __getitem__(self, key):
        return self.mglo.uniform(self.uniforms.get(key))

    def __setitem__(self, key, value):
        self.mglo.uniform(self.uniforms.get(key), value)

    def run(self, group_x=1, group_y=1, group_z=1):
        self.mglo.run(group_x, group_y, group_z)



NOTHING = 0
BLEND = 1
DEPTH_TEST = 2
CULL_FACE = 4
RASTERIZER_DISCARD = 8

NEAREST = 0x2600
LINEAR = 0x2601
NEAREST_MIPMAP_NEAREST = 0x2700
LINEAR_MIPMAP_NEAREST = 0x2701
NEAREST_MIPMAP_LINEAR = 0x2702
LINEAR_MIPMAP_LINEAR = 0x2703

POINTS = 0x0000
LINES = 0x0001
LINE_LOOP = 0x0002
LINE_STRIP = 0x0003
TRIANGLES = 0x0004
TRIANGLE_STRIP = 0x0005
TRIANGLE_FAN = 0x0006
LINES_ADJACENCY = 0x000A
LINE_STRIP_ADJACENCY = 0x000B
TRIANGLES_ADJACENCY = 0x000C
TRIANGLE_STRIP_ADJACENCY = 0x000D
PATCHES = 0x000E

CLAMP_TO_EDGE_X = 0x000001
CLAMP_TO_EDGE_Y = 0x000100
CLAMP_TO_EDGE_Z = 0x010000

REPEAT_X = 0x000002
REPEAT_Y = 0x000200
REPEAT_Z = 0x020000

MIRRORED_REPEAT_X = 0x000004
MIRRORED_REPEAT_Y = 0x000400
MIRRORED_REPEAT_Z = 0x040000

MIRROR_CLAMP_TO_EDGE_X = 0x000008
MIRROR_CLAMP_TO_EDGE_Y = 0x000800
MIRROR_CLAMP_TO_EDGE_Z = 0x080000

CLAMP_TO_BORDER_X = 0x000010
CLAMP_TO_BORDER_Y = 0x001000
CLAMP_TO_BORDER_Z = 0x100000



from typing import Any, Dict

from .attribute import Attribute
from .uniform import Uniform


class Program:
    __slots__ = ['mglo', 'attributes', 'uniforms', 'extra', 'old']

    def __init__(self, mglo, attributes, uniforms):
        self.mglo = mglo  # type: Any
        self.attributes = attributes  # type: Dict[str, Attribute]
        self.uniforms = uniforms  # type: Dict[str, Uniform]
        self.extra = None  # type: Any

    def __getitem__(self, key):
        return self.mglo.uniform(self.uniforms.get(key))

    def __setitem__(self, key, value):
        self.mglo.uniform(self.uniforms.get(key), value)



from typing import Any

from .buffer import Buffer
from .program import Program
from .scope import Scope


class VertexArray:
    __slots__ = ['mglo', 'program', 'scope', 'mode', 'vertices', 'extra', 'old']

    def __init__(self, mglo, program, scope, mode, vertices):
        self.mglo = mglo  # type: Any
        self.program = program  # type: Program
        self.scope = scope  # type: Scope
        self.mode = mode  # type: int
        self.vertices = vertices  # type: int
        self.extra = None  # type: Any

    def render(self, mode=None, vertices=-1, first=0, instances=1, color_mask=0xffffffffffffffff, depth_mask=1):
        self.mglo.render(mode, vertices, first, instances, color_mask, depth_mask)

    def render_indirect(self, buffer, mode=None, count=-1, first=0, color_mask=0xffffffffffffffff, depth_mask=1):
        self.mglo.render_indirect(mode, buffer, mode, count, first, color_mask, depth_mask)

    def transform(self, output, mode=None, vertices=-1, first=0, instances=1, flush=True):
        self.mglo.transform(output, mode, vertices, first, instances, flush)

    def transform_indirect(self, output, buffer, mode=None, count=-1, first=0, flush=True):
        self.mglo.transform_indirect(output, buffer, mode, count, first, flush)

    @property
    def index_buffer(self):
        return self.mglo.ibo

    @index_buffer.setter
    def index_buffer(self, value):
        self.mglo.ibo = value



from typing import Any, Tuple, Union

from .texture import Texture


class Sampler:
    __slots__ = ['mglo', 'extra']

    def __init__(self, mglo):
        self.mglo = mglo  # type: Any
        self.extra = None  # type: Any

    @property
    def filter(self) -> Tuple[int, int]:
        return self.mglo.filter

    @filter.setter
    def filter(self, value):
        self.mglo.filter = value

    @property
    def wrap(self) -> int:
        return self.mglo.wrap

    @wrap.setter
    def wrap(self, value):
        self.mglo.wrap = value

    @property
    def anisotropy(self) -> float:
        return self.mglo.anisotropy

    @anisotropy.setter
    def anisotropy(self, value):
        self.mglo.anisotropy = value

    @property
    def compare_func(self) -> str:
        return self.mglo.compare_func

    @compare_func.setter
    def compare_func(self, value):
        self.mglo.compare_func = value

    @property
    def lod_range(self) -> Tuple[int, int]:
        return self.mglo.lod_range

    @lod_range.setter
    def lod_range(self, value):
        self.mglo.lod_range = value

    @property
    def lod_bias(self) -> float:
        return self.mglo.lod_bias

    @lod_bias.setter
    def lod_bias(self, value):
        self.mglo.lod_bias = value

    @property
    def border(self) -> Tuple[float, float, float, float]:
        return self.mglo.border

    @border.setter
    def border(self, value):
        self.mglo.border = value

    @property
    def texture(self) -> Texture:
        return self.mglo.texture

    @texture.setter
    def texture(self, value):
        self.mglo.texture = value

    def use(self, location=0) -> None:
        self.mglo.use(location)



from typing import Any, Tuple


class Texture:
    __slots__ = ['mglo', '_level', '_layer', 'size', 'extra', 'old']

    def __init__(self, mglo, level, layer, size):
        self.mglo = mglo  # type: Any
        self._level = level  # type: int
        self._layer = layer  # type: int
        self.size = size  # type: Union[Tuple[int, int], Tuple[int, int, int]]
        self.extra = None  # type: Any

    @property
    def swizzle(self) -> str:
        return self.mglo._swizzle

    @swizzle.setter
    def swizzle(self, value):
        self.mglo.swizzle = value

    def level(self, level) -> 'Texture':
        return self.mglo.sub(level, self._layer)

    def layer(self, layer) -> 'Texture':
        return self.mglo.sub(self._level, layer)

    def write(self, data, viewport=None, alignment=1):
        self.mglo.write(data, viewport, alignment, self._level)

    def read(self, alignment=1, np=False) -> bytes:
        return self.mglo.read(self._level, self._layer, alignment, np)

    def build_mipmaps(self, base=0, max=-1):
        self.mglo.build_mipmaps(base, max)

    def bind_to_image(self, binding, access, format):
        self.mglo.bind(binding, access, format)



from typing import Any


class Buffer:
    __slots__ = ['mglo', 'size', 'extra', 'old']

    def __init__(self, mglo, size):
        self.mglo = mglo  # type: Any
        self.size = size  # type: int
        self.extra = None  # type: Any

    def write(self, data, offset=0):
        self.mglo.write(data, offset)

    def read(self, size=-1, offset=0, dtype=None) -> bytes:
        return self.mglo.read(size, offset, dtype)

    def read_into(self, buffer, size=-1, offset=0, write_offset=0):
        write_end = write_offset + (self.size if size < 0 else size)
        memoryview(buffer)[write_offset:write_end] = self.mglo.map(size, offset, True, False, None)
        self.mglo.unmap()

    def map(self, size=-1, offset=0, readable=False, writable=False, dtype=None) -> memoryview:
        return self.mglo.map(size, offset, readable, writable, dtype)

    def unmap(self):
        self.mglo.unmap()

    def clear(self):
        self.mglo.clear()

    def bind_to_uniform_block(self, binding=0, offset=0, size=-1):
        self.mglo.bind(binding, offset, size, 0)

    def bind_to_storage_buffer(self, binding=0, offset=0, size=-1):
        self.mglo.bind(binding, offset, size, 1)



class Limits:
    __slots__ = [
        'point_size_range',
        'point_size_granularity',
        'max_texture_size',
        'max_viewport_dims',
        'subpixel_bits',
        'max_3d_texture_size',
        'max_elements_vertices',
        'max_elements_indices',
        'smooth_point_size_range',
        'smooth_point_size_granularity',
        'smooth_line_width_range',
        'smooth_line_width_granularity',
        'aliased_line_width_range',
        'max_cube_map_texture_size',
        'point_fade_threshold_size',
        'max_texture_lod_bias',
        'max_draw_buffers',
        'max_vertex_attribs',
        'max_texture_image_units',
        'max_fragment_uniform_components',
        'max_vertex_uniform_components',
        'max_varying_floats',
        'max_combined_texture_image_units',
        'max_vertex_texture_image_units',
        'max_clip_distances',
        'major_version',
        'minor_version',
        'max_array_texture_layers',
        'min_program_texel_offset',
        'max_program_texel_offset',
        'max_varying_components',
        'max_renderbuffer_size',
        'max_color_attachments',
        'max_samples',
        'max_texture_buffer_size',
        'max_rectangle_texture_size',
        'max_vertex_uniform_blocks',
        'max_geometry_uniform_blocks',
        'max_fragment_uniform_blocks',
        'max_combined_uniform_blocks',
        'max_uniform_buffer_bindings',
        'max_uniform_block_size',
        'max_combined_vertex_uniform_components',
        'max_combined_geometry_uniform_components',
        'max_combined_fragment_uniform_components',
        'uniform_buffer_offset_alignment',
        'max_geometry_texture_image_units',
        'max_geometry_uniform_components',
        'max_geometry_output_vertices',
        'max_geometry_total_output_components',
        'max_vertex_output_components',
        'max_geometry_input_components',
        'max_geometry_output_components',
        'max_fragment_input_components',
        'context_profile_mask',
        'max_server_wait_timeout',
        'max_sample_mask_words',
        'max_color_texture_samples',
        'max_depth_texture_samples',
        'max_integer_samples',
        'max_dual_source_draw_buffers',
        'max_vertex_uniform_vectors',
        'max_varying_vectors',
        'max_fragment_uniform_vectors',
        'max_viewports',
        'viewport_subpixel_bits',
        'viewport_bounds_range',
        'min_map_buffer_alignment',
        'max_vertex_atomic_counter_buffers',
        'max_tess_control_atomic_counter_buffers',
        'max_tess_evaluation_atomic_counter_buffers',
        'max_geometry_atomic_counter_buffers',
        'max_fragment_atomic_counter_buffers',
        'max_combined_atomic_counter_buffers',
        'max_vertex_atomic_counters',
        'max_tess_control_atomic_counters',
        'max_tess_evaluation_atomic_counters',
        'max_geometry_atomic_counters',
        'max_fragment_atomic_counters',
        'max_combined_atomic_counters',
        'max_atomic_counter_buffer_size',
        'max_atomic_counter_buffer_bindings',
        'max_image_units',
        'max_combined_image_units_and_fragment_outputs',
        'max_image_samples',
        'max_vertex_image_uniforms',
        'max_tess_control_image_uniforms',
        'max_tess_evaluation_image_uniforms',
        'max_geometry_image_uniforms',
        'max_fragment_image_uniforms',
        'max_combined_image_uniforms',
        'max_element_index',
        'max_compute_uniform_blocks',
        'max_compute_texture_image_units',
        'max_compute_image_uniforms',
        'max_compute_shared_memory_size',
        'max_compute_uniform_components',
        'max_compute_atomic_counter_buffers',
        'max_compute_atomic_counters',
        'max_combined_compute_uniform_components',
        'max_compute_work_group_invocations',
        'max_debug_message_length',
        'max_debug_logged_messages',
        'max_debug_group_stack_depth',
        'max_label_length',
        'max_uniform_locations',
        'max_framebuffer_width',
        'max_framebuffer_height',
        'max_framebuffer_layers',
        'max_framebuffer_samples',
        'max_num_active_variables',
        'max_num_compatible_subroutines',
        'max_vertex_shader_storage_blocks',
        'max_geometry_shader_storage_blocks',
        'max_tess_control_shader_storage_blocks',
        'max_tess_evaluation_shader_storage_blocks',
        'max_fragment_shader_storage_blocks',
        'max_compute_shader_storage_blocks',
        'max_combined_shader_storage_blocks',
        'max_shader_storage_buffer_bindings',
        'max_shader_storage_block_size',
        'max_combined_shader_output_resources',
        'max_vertex_attrib_relative_offset',
        'max_vertex_attrib_bindings',
        'max_vertex_attrib_stride',
        'max_cull_distances',
        'max_combined_clip_and_cull_distances',
        'max_texture_max_anisotropy',
    ]

    def json(self):
        res = {}
        for slot in Limits.__slots__:
            if hasattr(self, slot):
                res[slot] = getattr(self, slot)
        return res



from .attribute import Attribute
from .buffer import Buffer
from .compute_shader import ComputeShader
from .context import Context, create_context, extensions, hwinfo, inspect, glprocs, release
from .error import Error
from .framebuffer import Framebuffer
from .limits import Limits
from .program import Program
from .query import Query
from .renderbuffer import Renderbuffer
from .sampler import Sampler
from .scope import Scope
from .texture import Texture
from .uniform import Uniform
from .vertex_array import VertexArray
from .tools import compiler_error, linker_error, split_format

from .constants import (
    NOTHING, BLEND, DEPTH_TEST, CULL_FACE, RASTERIZER_DISCARD, NEAREST, LINEAR,
    NEAREST_MIPMAP_NEAREST, LINEAR_MIPMAP_NEAREST, NEAREST_MIPMAP_LINEAR, LINEAR_MIPMAP_LINEAR,
    POINTS, LINES, LINE_LOOP, LINE_STRIP, TRIANGLES, TRIANGLE_STRIP, TRIANGLE_FAN, LINES_ADJACENCY,
    LINE_STRIP_ADJACENCY, TRIANGLES_ADJACENCY, TRIANGLE_STRIP_ADJACENCY, PATCHES,
    CLAMP_TO_EDGE_X, CLAMP_TO_EDGE_Y, CLAMP_TO_EDGE_Z, REPEAT_X, REPEAT_Y, REPEAT_Z,
    MIRRORED_REPEAT_X, MIRRORED_REPEAT_Y, MIRRORED_REPEAT_Z, MIRROR_CLAMP_TO_EDGE_X,
    MIRROR_CLAMP_TO_EDGE_Y, MIRROR_CLAMP_TO_EDGE_Z, CLAMP_TO_BORDER_X,
    CLAMP_TO_BORDER_Y, CLAMP_TO_BORDER_Z,
)

__version__ = '6.0.0'

# TODO: remove
class Refholder:
    __slots__ = ['mglo', 'old']
    def __init__(self, mglo):
        self.mglo = mglo



from typing import Any


class Renderbuffer:
    __slots__ = ['mglo', 'size', 'extra', 'old']

    def __init__(self, mglo, size):
        self.mglo = mglo  # type: Any
        self.size = size  # type: Tuple[int, int]
        self.extra = None  # type: Any



from typing import Any

from .buffer import Buffer
from .compute_shader import ComputeShader
from .constants import NEAREST
from .framebuffer import Framebuffer
from .limits import Limits
from .program import Program
from .query import Query
from .renderbuffer import Renderbuffer
from .sampler import Sampler
from .scope import Scope
from .texture import Texture
from .vertex_array import VertexArray


def _attr_fmt(attr):
    return '%d%c' % (attr.cols * attr.rows * attr.size, attr.shape)


class Context:
    __slots__ = ['mglo', 'version_code', 'limits', 'screen', 'fbo', 'recorder', 'extra', 'old']

    def __init__(self, mglo, version_code, limits, screen, fbo, recorder):
        self.mglo = mglo  # type: Any
        self.version_code = version_code  # type: int
        self.limits = limits  # type: Limits
        self.screen = screen  # type: Framebuffer
        self.fbo = fbo  # type: Framebuffer
        self.recorder = recorder  # type: Any
        self.extra = None  # type: Any

    def buffer(self, data=None, reserve=0, readable=True, writable=True, local=False) -> Buffer:
        return self.mglo.buffer(data, reserve, readable, writable, local)

    def compute_shader(self, source) -> ComputeShader:
        return self.mglo.compute_shader(source)

    def program(self, vertex_shader, fragment_shader=None, geometry_shader=None, tess_control_shader=None, tess_evaluation_shader=None, varyings=()) -> Program:
        return self.mglo.program(vertex_shader, fragment_shader, geometry_shader, tess_control_shader, tess_evaluation_shader, varyings)

    def texture(self, size, components=3, data=None, levels=-1, samples=0, aligment=1, dtype='f1') -> Texture:
        return self.mglo.texture(0, size, components, data, levels, samples, aligment, dtype)

    def depth_texture(self, size, data=None, levels=-1, samples=0, aligment=1, dtype='d3') -> Texture:
        return self.mglo.texture(0, size, 1, data, levels, samples, aligment, dtype)

    def texture_array(self, size, components=3, data=None, levels=-1, samples=0, aligment=1, dtype='f1') -> Texture:
        return self.mglo.texture(2, size, components, data, levels, samples, aligment, dtype)

    def depth_texture_array(self, size, data=None, levels=-1, samples=0, aligment=1, dtype='d3') -> Texture:
        return self.mglo.texture(2, size, 1, data, levels, samples, aligment, dtype)

    def texture_cube(self, size, components=3, data=None, levels=-1, samples=0, aligment=1, dtype='f1') -> Texture:
        return self.mglo.texture(3, size, components, data, levels, samples, aligment, dtype)

    def depth_texture_cube(self, size, data=None, levels=-1, samples=0, aligment=1, dtype='d3') -> Texture:
        return self.mglo.texture(3, size, 1, data, levels, samples, aligment, dtype)

    def sampler(self, texture, filter=NEAREST, wrap=0, compare_func=None, anisotropy=None, lod_range=None, lod_bias=None, border=None) -> Sampler:
        return self.mglo.sampler(texture, filter, wrap, compare_func, anisotropy, lod_range, lod_bias, border)

    def vertex_array(self, program, content, index_buffer=None) -> VertexArray:
        return self.mglo.vertex_array(program, content, index_buffer)

    def simple_vertex_array(self, program, buffer, *attributes, index_buffer=None) -> VertexArray:
        content = [(buffer, ' '.join(_attr_fmt(program.attributes[a]) for a in attributes)) + attributes]
        return self.mglo.vertex_array(program, content, index_buffer)

    def scope(self, enable_only=-1, framebuffer=None, samplers=None, uniform_buffers=None, storage_buffers=None) -> Scope:
        return self.mglo.scope(enable_only, framebuffer, samplers, uniform_buffers, storage_buffers)

    def query(self, time=False, primitives=False, samples=False, any_samples=False) -> Query:
        return self.mglo.query(time, primitives, samples, any_samples)

    def renderbuffer(self, size, components=4, samples=0, dtype='f1') -> Renderbuffer:
        return self.mglo.renderbuffer(size, components, samples, dtype)

    def depth_renderbuffer(self, size, samples=0, dtype='d3') -> Renderbuffer:
        return self.mglo.renderbuffer(size, 1, samples, dtype)

    def framebuffer(self, color_attachments=(), depth_attachment=None) -> Framebuffer:
        return self.mglo.framebuffer(color_attachments, depth_attachment)

    def configure(self, key, **params):
        return self.mglo.configure(params)

    def clear(self, red=1.0, green=1.0, blue=1.0, alpha=1.0, depth=1.0, viewport=None):
        self.fbo.clear(0, (red, green, blue, alpha), viewport)
        self.fbo.clear(-1, depth, viewport)

    def copy_buffer(self, dst, src, size=-1, *, read_offset=0, write_offset=0) -> None:
        self.mglo.copy_buffer(dst, src, size, read_offset, write_offset)

    def record(self):
        return self.mglo.record()

    def replay(self, bytecode):
        return self.mglo.replay(bytecode)

    def make_current(self):
        self.mglo.make_current()


def create_context(standalone=False, require=None, glhook=None):
    import moderngl.mgl as mgl
    return mgl.new.create_context(standalone, require, glhook)


def extensions(context):
    import moderngl.mgl as mgl
    return mgl.new.extensions(context)


def hwinfo(context):
    import moderngl.mgl as mgl
    return mgl.new.hwinfo(context)


def glprocs(context):
    import moderngl.mgl as mgl
    return mgl.new.glprocs(context)


def release(obj):
    import moderngl.mgl as mgl
    return mgl.new.release(obj)


def inspect(obj):
    import moderngl.mgl as mgl
    return mgl.new.inspect(obj)



from typing import Any


class Query:
    __slots__ = ['mglo', 'elapsed', 'primitives', 'samples', 'extra']

    def __init__(self, mglo):
        self.mglo = mglo  # type: Any
        self.elapsed = None  # type: int
        self.primitives = None  # type: int
        self.samples = None  # type: int
        self.extra = None  # type: Any

    def __enter__(self):
        self.mglo.begin()
        return self

    def __exit__(self, *args):
        self.mglo.end()



from typing import Any


class Scope:
    __slots__ = ['mglo', 'extra']

    def __init__(self, mglo):
        self.mglo = mglo  # type: Any
        self.extra = None  # type: Any

    def __enter__(self):
        self.mglo.begin()

    def __exit__(self, *args):
        self.mglo.end()



class Uniform:
    __slots__ = ['_type', 'location', 'cols', 'rows', 'size', 'shape']

    def __init__(self, _type, location, cols, rows, size, shape):
        self._type = _type  # type: int
        self.location = location  # type: int
        self.cols = cols  # type: int
        self.rows = rows  # type: int
        self.size = size  # type: int
        self.shape = shape  # type: int



import textwrap
import re

from .error import Error


def number_source(source):
    lines = textwrap.dedent(source).split('\n')
    fmt = ' %%%dd | %%s' % len(str(len(lines)))
    return '\n'.join(fmt % (i + 1, line) for i, line in enumerate(lines))


def compiler_error(name, source, info):
    message = 'compiler error (%s)\n\n%s\n\n%s' % (name, number_source(source), info)
    raise Error(message)


def linker_error(info):
    raise Error(info)


def split_format(fmt):
    res = []
    divisor = 0
    stride = 0
    nodes = fmt.replace('/', '  /').split()
    if nodes[-1].startswith('/'):
        match = re.match(r'^/(?:(\d*)i|([vr]))$', nodes[-1])
        if not match:
            raise Error('%r is not a valid divisor' % nodes[-1])
        a, b = match.groups()
        divisor = int(a) if a else 1 if not b else 0 if b == 'v' else 0x7fffffff
        nodes = nodes[:-1]
    for node in nodes:
        match = re.match(r'^(\d*)([fiux])([1248]?)$', node)
        if not match:
            raise Error('%r is not a valid format node' % node)
        a, b, c = match.groups()
        a = int(a) if a else 1
        c = int(c) if c else 1 if b == 'x' else 4
        res.append((a, b, c))
        stride += a * c
    return res, divisor, stride
